/*
 *   EPRSD Intra and Inter task memory trace compression program
 *   Single Author info:
 *   sbudanu SANDEEP BUDANUR RAMANNA
 *
 */
#include "mpi.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "myprsdtlib.h"

#define PRSD_PIPE "/home/sbudanu/pin-2.5-24110-gcc.4.0.0-ia32_intel64-linux/source/tools/MyMemTrace/16x16_32T.txt"
//"8x8_4T.trace" //./sample.trace"        //"./STABLE/t1.txt" //"./8x8_4T.trace"
#define BUF_LEN 200 //should be greater than equal to size of any data we are sending using MPI_Send()
#define MSG_TAG 99
#define MAIN_PROCESS 0

FILE* fpipe = NULL;
FILE* outfile = NULL;
uint64_t sincelast = 0;
uint64_t ip, ebp, sp, sign, thread_id, addr;
char type = 'x', valid = 'N';
char rcvBuffer[BUF_LEN];


enum PRSD_COMP_MSGS_T
{
 //main process to child process messages, prefix MC_
 MC_TRACE, //msg having trace
 MC_TRACES_END, //msg indicating traces are over
 MC_RCV_FROM,//instruction to receive PRSDs from a specific process
 MC_SND_TO, //instructiion to send PRSDs to a specific process
 MC_SAVE_TO_FILE, //instruction to save PRSDs to a specified file
 MC_ERROR, //indicates ERROR
 MC_EXIT, //instruction for the child process to end
 MC_ACTIVE, //indicates that the main process is still active

 //child process to main process messages, prefix CM_
 CM_QUERY_STATUS, //msg to
 CM_READY_FOR_MERGE,
 CM_MERGE_DONE, //msg indicating the merge is completed with a specific process's PRSDs
 CM_PRSDS_SENT, //msg indiacting that all PRSDs were received
 CM_PRSDS_RCVD, //msg indicating that all PRSDs were sent

 //child process to child process messages, prefix CC_
 CC_PRSD, //msg having PRSD
 CC_PRSDS_END, //indiactes PRSDs are over
 MSGS_MAX
};

enum STATUS
{
  STATUS_NONE = 0,
  STATUS_PRSD_SNDR,
  STATUS_PRSD_RCVR,
  STATUS_PRSD_MERGING,
  STATUS_EXIT,
  STATUS_MAX
};

/* example of deriving a class from abstract class INFO*/
class MYINFO:public INFO
{
	public:
	MYINFO():INFO()
	{
	}

	MYINFO(MYINFO& src):INFO()
	{
		//cout<<"\nMYINFO copy ctor called";
		value = src.value;
	}

	~MYINFO()
	{
		//cout <<"\nMYINFO dtor()";
	}

	virtual INT64 getValue()
	{
		//cout<<"\nMYINFO getValue called";
		return value;
	}
	virtual void setValue(void* val)
	{
		//cout<<"\nMYINFO setValue called";
		value = (INT64)val;
	}

};


/* class defining match function */
template <class T>
class MYMATCH
{
	public:
	/* compare function for intra-task compression */
	inline static bool compare(T* src, T* dest)
	{
		//nothing to compare - return true
		//returning false makes all EPRSDs to be appended - uncompressed case
		return true;
	}

	/* compare function for inter-task compression */
	inline static bool intertask_compare(T* src, T* dest)
	{
		return true;
	}
};

/* merge function hook */
template <class T>
class MYMERGE
{
	public:
	/* template library merges the EPRSDs, this is just a notification to user - is it really required?*/
	/* EPRSDs are private to compressor object, cannot let user merge them. Library merges the EPRSDs after
	matching the signature and userdefined part in MYMATCH::compare()
	*/
	inline static void mergecb(T* src, T* dest)
	{
		//cout<<endl<<"Merge_cb Src:"<<src->getValue()<<", "<<"Dest:"<<dest->getValue();
		return;
	}

	inline static void intertask_mergecb(T* src, T* dest)
	{
		return;
	}
};

//pack the tarce info and send to remote process
void sendTraceMsg(int dest)
{
    char sndBuffer[BUF_LEN];
    int position = 0;
    int msg_type = MC_TRACE;

    MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&ip, 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&ebp, 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&sp, 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&addr, 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&sign, 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&thread_id, 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&sincelast, 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Pack(&type, 1, MPI_CHAR, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);
    //sleep(1);
    return;
}

void sendTracesEndMsg(int dest)
{
    char sndBuffer[BUF_LEN];
    int position = 0;
    int msg_type = MC_TRACES_END;
    MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);
    return;
}

void sendAllPRSDs(int dest, EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR *iter)
{
    char sndBuffer[BUF_LEN];
    int position = 0, i=0;
    int msg_type = CC_PRSD;
    INT64 value = 1122;
    INT16 code = 0;
    EPRSD_DATA<MYINFO> *cur_prsd = NULL;

    //printf("sendAllPRSDs - 1, dest: %d\n", dest);
    iter->begin();
    while(1)
    {
        //take individual PRSDs
        //printf("$dest: %d$\n", dest);
	cur_prsd = iter->getNext();
	if(!cur_prsd) break;
        position = 0;
        MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	code = 0xFFFF & cur_prsd->code;
        MPI_Pack(&code, 1, MPI_SHORT, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        MPI_Pack(&(cur_prsd->det_len), 1, MPI_SHORT, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);

	//details
	for(i=0;i<(cur_prsd->det_len);i++)
	{
        	MPI_Pack(&(cur_prsd->det[i].type), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		/* intra task info */
        	MPI_Pack(&(cur_prsd->det[i].start_value), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].length), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].stride), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		/* inter task info */
        	MPI_Pack(&(cur_prsd->det[i].tid_start_value), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].tid_length), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].tid_stride), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		/* inter node info */
        	MPI_Pack(&(cur_prsd->det[i].node_id_start_value), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].node_id_length), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].node_id_stride), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);

        	MPI_Pack(&(cur_prsd->det[i].loopsize), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].loopcount), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        	MPI_Pack(&(cur_prsd->det[i].eprsd_merge_count), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	}

	//common part
        MPI_Pack(&((cur_prsd->cmninfo)->max_sign_len), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        MPI_Pack(&((cur_prsd->cmninfo)->sign_len), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	for(i=0;i<(cur_prsd->cmninfo)->sign_len;i++)
	{
	        MPI_Pack(&((cur_prsd->cmninfo)->signature[i]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	}
        MPI_Pack(&((cur_prsd->cmninfo)->ref_id), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        MPI_Pack(&((cur_prsd->cmninfo)->type), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        MPI_Pack(&((cur_prsd->cmninfo)->thread_id), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        MPI_Pack(&((cur_prsd->cmninfo)->node_id), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
        MPI_Pack(((cur_prsd->cmninfo)->name), 10, MPI_CHAR, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);

	//user-defined part
	value = (cur_prsd->userinfo)->getValue();
        MPI_Pack(&value, 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);
    }
    printf("\n***All PRSDs sent to %d\n", dest);
    return;
}

void sendPRSDsEndMsg(int dest)
{
    char sndBuffer[BUF_LEN];
    int position = 0;
    int msg_type = CC_PRSDS_END;
    MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
    MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);
    return;
}

int M_SndMsgsProc(int args)
{
  int num = 0, dest = 0, n_processes = 0;
  int myrank = (int)args;
  FILE* fptr = NULL;
  COMMON* cmn = NULL;
  MYINFO* address = NULL;
  UINT64 *arr_sign = NULL;

  /* create a EPRSD compressor object of type MYINFO */
  EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE> ecr1;
  EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR iter(&ecr1);
  cmn = new COMMON(10);//sign len of 10 max
  address = new MYINFO();
  arr_sign = (UINT64*)malloc(10*sizeof(UINT64));

  printf("M_SndMsgsProc starting\n");
  MPI_Comm_size(MPI_COMM_WORLD, &n_processes);
  printf("no. of processes: %d\n", n_processes);

  while(1)
  {
   num = fscanf(fpipe,"ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u %c %c addr:%20p %u\n",
                     &ip,     &ebp,     &sp,     &sign,      &thread_id, &type, &valid, &addr, &sincelast);
   /*
   fprintf(outfile,"ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u %c %c addr:%20p %u\n",
                     ip,     ebp,     sp,     sign,      thread_id, type, valid, addr, sincelast);
   */

   if(num == 0 || num == -1)
   {
     printf("\nEOF reached or ERROR....%d\n",num);
     //to indicate traces are over
     for(dest=1;dest<n_processes;dest++)
     {
      sendTracesEndMsg(dest);
     }

     //save main process's PRSDs to a file
     #if 0 //27 Mar 2009
     sprintf(filepath, "%d.prsd", myrank);
     if((fptr = fopen(filepath,"w")) == NULL)
     {
      printf("%s - File Creation Error\n", filepath);
      return -1;
     }
     prsdCompressor[0].saveToFile(fptr);
     fclose(fptr);
     #endif
     //sendAllPRSDs(n_processes-1);//process with highest rank
     break;
   }

   //if fscanf was successful in reading the trace
   if(thread_id > 0 && thread_id < n_processes)
   {
    //send trace to other processes for compression
    sendTraceMsg((int)thread_id);
   }
   else if(thread_id == 0)
   {
    //compress in this(main) process only, add to compressor object
    cmn->setData(NULL, 0, sign, type,
                 0x00000000FFFFFFFF & thread_id,
		 0x00000000FFFFFFFF & 0,
		 "LD_ST");
    address->setValue((void*)addr);
    ecr1.add(cmn, address);
   }
   else
   {
     printf("Invalid thread id: %d ...exiting\n", thread_id);
     //to indicate traces are over
     for(dest=1;dest<n_processes;dest++)
     {
      sendTracesEndMsg(dest);
      //MPI_Send(&gs_trace, sizeof(gs_trace), MPI_BYTE, dest, MSG_TAG, MPI_COMM_WORLD);
     }
     break;
   }
  }//while ends

  ecr1.finalize();
  fprintf(outfile, "\nAfter finalize");
  ecr1.print(outfile);
  printf("\n");

  return 0;
}

//Main process creates two threads
void startMainProcess(int myrank)
{
  int n_processes = -1;
  char outstr[16];
  MPI_Comm_size(MPI_COMM_WORLD, &n_processes);
  printf("no. of processes: %d\n", n_processes);
  printf("ROOT %d starting...\n", myrank);

  if((fpipe = fopen(PRSD_PIPE, "r"))== NULL)
  {
   printf("pipe %s not present....:!exiting\n", PRSD_PIPE);
   return;
  }

  sprintf(outstr, "%d.txt", myrank);
  if((outfile = fopen(outstr, "w")) == NULL)
  {
   printf("outfile open error in %d process....exiting\n", myrank);
   return;
  }

  M_SndMsgsProc(myrank);
  return;
}

void startChildProcess(int myrank)
{
  MPI_Status status;
  char outstr[16];
  FILE* fptr = NULL;
  int position = 0, traces_over = 0;
  int msg_type = 0;
  //int sender = MAIN_PROCESS; //default
  //int receiver = MAIN_PROCESS; //default
  int n_processes = -1, p=0, i=0, stride = 0;
  COMMON* cmn = NULL;
  MYINFO* address = NULL;
  UINT64 *arr_sign = NULL;

  /* create a EPRSD compressor object of type MYINFO */
  EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE> ecr1;
  EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR iter(&ecr1);
  cmn = new COMMON(10);//sign len of 10 max
  address = new MYINFO();
  arr_sign = (UINT64*)malloc(10*sizeof(UINT64));

  MPI_Comm_size(MPI_COMM_WORLD, &n_processes);
  printf("no. of processes: %d\n", n_processes);
  stride = n_processes;

  printf("CHILD %d starting...\n", myrank);

  sprintf(outstr, "%d.txt", myrank);
  if((outfile = fopen(outstr, "w")) == NULL)
  {
   printf("outfile open error in %d process....exiting\n", myrank);
   return;
  }

	//Receive the messages from ROOT process
	while(1)
  	{
	   MPI_Recv(rcvBuffer,BUF_LEN, MPI_PACKED,
            MPI_ANY_SOURCE, //sender, //src PID
            MSG_TAG, MPI_COMM_WORLD, &status);

	   //unpack the buffer contents
	   position = 0;
	   MPI_Unpack(rcvBuffer, BUF_LEN, &position, &msg_type, 1, MPI_UNSIGNED, MPI_COMM_WORLD);

	   switch(msg_type)
	   {
	     case MC_TRACE:
	     //printf("CHILD %d trace rcvd\n", myrank);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &ip, 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &ebp, 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &sp, 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &addr, 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &sign, 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &thread_id, 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &sincelast, 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	     MPI_Unpack(rcvBuffer, BUF_LEN, &position, &type, 1, MPI_CHAR, MPI_COMM_WORLD);
    	     cmn->setData(NULL, 0, sign, type,
                 0x00000000FFFFFFFF & thread_id,
		 0x00000000FFFFFFFF & 0,
		 "LD_ST");
	     address->setValue((void*)addr);
	     ecr1.add(cmn, address);
     	     break;

	     case CC_PRSD:
	     {
		int i = 0;
		CODE code;
		INT16 det_len;
		INT64 value = 0;
		struct details det[3];
		EPRSD_DATA<MYINFO> tmpdata;
	     	MPI_Unpack(rcvBuffer, BUF_LEN, &position, &code, 1, MPI_SHORT, MPI_COMM_WORLD);
	     	MPI_Unpack(rcvBuffer, BUF_LEN, &position, &det_len, 1, MPI_SHORT, MPI_COMM_WORLD);
		printf("\n%d received PRSD: code = %d, det_len = %d", myrank, code, det_len);
		//details
		for(i=0;i<det_len && i<3;i++)
		{
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].type), 1, MPI_LONG, MPI_COMM_WORLD);
			/* intra task info */
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].start_value), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].length), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].stride), 1, MPI_LONG, MPI_COMM_WORLD);
			/* inter task info */
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_start_value), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_length), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_stride), 1, MPI_LONG, MPI_COMM_WORLD);
			/* inter node info */
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].node_id_start_value), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].node_id_length), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].node_id_stride), 1, MPI_LONG, MPI_COMM_WORLD);

			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].loopsize), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].loopcount), 1, MPI_LONG, MPI_COMM_WORLD);
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].eprsd_merge_count), 1, MPI_LONG, MPI_COMM_WORLD);
			//printf(" start_value = %p, length = %d, stride = %d, loopsize = %d, loopcount = %d, emc = %d", \
			det[i].start_value, det[i].length, det[i].stride, det[i].loopsize, det[i].loopcount, det[i].eprsd_merge_count);
		}
		//common part
	        MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->max_sign_len), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
	        MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->sign_len), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);

		for(i=0;i<cmn->sign_len;i++)
	        {
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->signature[i]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
		}
		MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->ref_id), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
		MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->type), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
		MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->thread_id), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
		MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->node_id), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
		MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->name), 10, MPI_CHAR, MPI_COMM_WORLD);

		//user-defined part
		MPI_Unpack(rcvBuffer, BUF_LEN, &position, &value, 1, MPI_LONG, MPI_COMM_WORLD);
		address->setValue((void*)value);

		//printf(" sign_len = %d", cmn->sign_len);
		//printf(" value = %p", value);
		//printf(" ref_id = %p", cmn->ref_id);
		//printf(" type = %d", cmn->type);
		//printf(" tid = %d", cmn->thread_id);
		//printf(" nid = %d", cmn->node_id);
		//printf(" name = %s", cmn->name);
		tmpdata.copy(code, det, det_len, cmn, address);
		ecr1.add(&tmpdata);
	     }
	     //receive from myrank+stride process
	     break;

	     case CC_PRSDS_END:
	     {
		printf("\nCHILD %d traces/prsds end\n", myrank);
		stride = stride/2;
		if(myrank > stride)
		{
			if(stride > 0)
			{
			 printf("\nSENDER %d - RECEIVER %d", myrank, myrank-stride);
			 //send all EPRSDs
			 sendAllPRSDs(myrank-stride, &iter);
			 //notify the end
			 sendPRSDsEndMsg(myrank-stride);
			}
			else
			{
			 printf("%d dumping EPRSDs", myrank);
			 ecr1.print(outfile); //process 1 dumps the reduced EPRSDs
			 //printf("\nSENDER %d - RECEIVER %d", myrank, 0);
			 //sendPRSDsEndMsg(0);
			}
			return;
		}
		else
		{
			printf("\nRECEIVER %d - SENDER %d", myrank, myrank+stride);

		}
		//ecr1.finalize();
  		//fprintf(outfile, "\nAfter finalize");
		//ecr1.print(outfile);
		//return;
	     }
	     break;

	     case MC_TRACES_END:
	     {
		printf("\nCHILD %d traces/prsds end\n", myrank);
		ecr1.finalize();
		stride = stride/2;
		if(myrank > stride)
		{
			if(stride > 0)
			{
			 printf("\nSENDER %d - RECEIVER %d", myrank, myrank-stride);
			 //send all EPRSDs
			 sendAllPRSDs(myrank-stride, &iter);
			 //notify the end
			 sendPRSDsEndMsg(myrank-stride);
			}
			else
			{
			 ecr1.print(outfile); //process 0 dumps the reduced EPRSDs
			 //printf("\nSENDER %d - RECEIVER %d", myrank, 0);
			 //sendPRSDsEndMsg(0);
			}
			//send all EPRSDs
			//notify the end
			return;
		}
		else
		{
			printf("\nRECEIVER %d - SENDER %d", myrank, myrank+stride);

		}
		//ecr1.finalize();
  		//fprintf(outfile, "\nAfter finalize");
		//ecr1.print(outfile);
		//return;
	     }
  	     break;

	     default:
		printf("Unknown msg type %d received\n", msg_type);
	     break;
	  }//switch ends
	}//while ends
  return;
}

//starting point of the program
int main(int argc, char** argv)
{
 int myrank = 0, provided = 0;
 //MPI_Init( &argc, &argv );
 MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);

 if(provided == MPI_THREAD_MULTIPLE)
 {
  //printf("MPI threads supported\n");
 }

 MPI_Comm_rank( MPI_COMM_WORLD, &myrank );

 if (myrank == 0) /* code for process zero */
 {
   startMainProcess(myrank);
 }
 else /* code for child process */
 {
   startChildProcess(myrank);
 }

 MPI_Barrier(MPI_COMM_WORLD);
 MPI_Finalize();
 return 0;
}
